from mpl_toolkits.mplot3d import Axes3D
import tkinter as tk
from tkinter import ttk
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
from matplotlib.figure import Figure
import matplotlib.pyplot as plt
import numpy as np

import cv2


class IndexGUI(object):
    ref_dump_path = None

    def __init__(self, file_path):
        self.pxpy = np.array(np.loadtxt(file_path, dtype=str, delimiter=',')[1:, :], dtype=float)
        self.source_index = 0

        self.root = tk.Tk()
        self.root.title('Index Finder')

        # distances related
        self.distances = None
        self.dis_image = None

        # index label
        o = tk.Label(self.root, text="Index:", width=10)
        o.grid(row=1, column=0)

        # index
        variable = tk.IntVar(0)
        l = tk.Entry(self.root, textvariable=variable)
        l.grid(row=1, column=1)
        scale = tk.Scale(self.root, orient=tk.HORIZONTAL, length=300,
                              from_=0, to=len(self.pxpy)-1, resolution=1, variable=variable, command=self.update_index)
        scale.grid(row=1, column=2, columnspan=2, padx=10)
        l.bind('<Return>', self.update_index)

        self.scale = scale
        self.index_var = variable

        # distance label
        o = tk.Label(self.root, text="distance:", width=10)
        o.grid(row=2, column=0)

        # distance vald
        variable = tk.DoubleVar(0)
        l = tk.Label(self.root, textvariable=variable)
        l.grid(row=2, column=1)
        self.distance_var = variable

        # save ref button
        variable = tk.DoubleVar(0)
        o = tk.Button(self.root, text='Save As Ref', command=self.save_trace_as_ref)
        o.grid(row=2, column=2)

        # Graph
        f = Figure(figsize=(5, 4), dpi=100)
        a = f.add_subplot(111)

        frame = ttk.Frame(self.root)
        frame.grid(row=3, column=0, columnspan=4)

        canvas = FigureCanvasTkAgg(f, master=frame)
        canvas.draw()
        canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=True)

        toolbar = NavigationToolbar2Tk(canvas, frame)
        toolbar.update()
        canvas._tkcanvas.pack(side=tk.TOP, fill=tk.BOTH, expand=True)
        self.canvas = canvas
        self.axis = a
        self.axis.plot(self.pxpy[:, 0], self.pxpy[:, 1], 'ob', markersize=1)

        # image component
        self.img_canvas = None
        self.img_axis = None
        self.row_var = None
        self.column_var = None

        # set general config
        self.root.grid_rowconfigure(1, minsize=100)

        # draw plot with initial settings
        self.update_ui()

    def init_image_component(self):
        # row variable
        l = tk.Label(self.root, text='Row: ')
        l.grid(row=1, column=4)
        variable = tk.IntVar(0)
        l = tk.Label(self.root, textvariable=variable)
        l.grid(row=1, column=5)
        self.row_var = variable

        # col variable
        l = tk.Label(self.root, text='Column: ')
        l.grid(row=1, column=6)
        variable = tk.IntVar(0)
        l = tk.Label(self.root, textvariable=variable)
        l.grid(row=1, column=7)
        self.column_var = variable

        # Image plot
        f = Figure(figsize=(5, 4), dpi=100)
        a = f.add_subplot(111)

        frame = ttk.Frame(self.root)
        frame.grid(row=3, column=4, columnspan=4)

        canvas = FigureCanvasTkAgg(f, master=frame)
        canvas.draw()
        canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=True)

        toolbar = NavigationToolbar2Tk(canvas, frame)
        toolbar.update()
        canvas._tkcanvas.pack(side=tk.TOP, fill=tk.BOTH, expand=True)
        self.img_canvas = canvas
        self.img_axis = a

    def update_index(self, event):
        self.update_ui()

    def calc_distance(self, trace, ref, offset):
        sig = trace - ref
        d_i = np.argmax(trace - ref)

        if sig[d_i] > 100:
            d = (d_i - offset) * self.lr.sample_distance
        else:
            d = 0
            d_i = offset

        return d_i, d, sig

    def build_distances_image(self):
        distances = np.empty(len(self.traces))
        for i, t in enumerate(self.traces):
            _, d, _ = self.calc_distance(t, self.ref, self.source_index.start)
            distances[i] = d

        if self.distance_low_thr:
            distances[distances < self.distance_low_thr] = 0
        distances[distances < 0] = 0

        self.distances = distances
        dis_image = self.lr.get_2d_image(distances)
        self.dis_image = np.stack((dis_image,)*3, -1)

    def save_trace_as_ref(self):
        index = self.index_var.get()
        self.ref = self.pxpy[index, :]
        np.save(IndexGUI.ref_dump_path, self.ref)
        self.build_distances_image()
        self.update_ui()

    def update_ui(self):
        index = self.index_var.get()
        trace = self.pxpy[index, :]
        ind = np.arange(len(trace))

        # update plot
        self.axis.clear()
        self.axis.plot(self.pxpy[:, 0], self.pxpy[:, 1], 'ob', markersize=1)
        self.axis.plot(self.pxpy[index, 0], self.pxpy[index, 1], 'or')

        # self.axis.set_xlim(-max(angles), max(angles))
        # self.axis.set_ylim(-max(angles), max(angles))
        self.canvas.draw()

        # update image
        # if self.is_image:
        #     # update row, column labels
        #     r, c = self.lr.img_index_to_r_c(self.index_var.get())
        #     self.row_var.set(r)
        #     self.column_var.set(c)
        #
        #     # update image
        #     # max_ = np.max(self.dis_image)
        #     # img = np.array(self.dis_image / max_ * 255, np.uint8)
        #     # cv2.circle(img, (c, r), 4, color=(255, 0, 0), thickness=2)
        #     self.img_axis.clear()
        #     self.img_axis.imshow(self.dis_image[:, :, 0], cmap='gray')
        #
        #     self.img_canvas.draw()




from ctypes import *
import numpy as np
import os
from find_centers import find_center


class PylonCamera:
    def __init__(self, settings_path=None):
        self.is_init = False
        dir = os.path.dirname(__file__)
        dll_path = dir + r'/x64/Release/PylonCameraDll.dll'
        self.camdll = cdll.LoadLibrary(dll_path)

        # Terminate
        # Create Video
        self.camdll.CreateVideo.restype = c_int
        self.camdll.CreateVideo.argtypes = [c_char_p, c_int, c_int, POINTER(c_int)]

        # Grab image
        self.camdll.GetImage.restype = c_int
        self.camdll.GetImage.argtypes = [POINTER(c_ubyte), c_int, c_int]

        # Gram images sequence
        self.camdll.GrabImagesSequence.restype = c_int
        self.camdll.GrabImagesSequence.args = [c_int, POINTER(POINTER(c_ubyte))]

        # Get Camera Info
        self.camdll.GetCameraInfo.restype = c_int
        self.camdll.GetCameraInfo.argtypes = [POINTER(c_int), POINTER(c_int), POINTER(c_float), POINTER(c_float)]

        # Save camera settings
        self.camdll.SaveCameraSettings.restype = c_int
        self.camdll.SaveCameraSettings.argtypes = [c_char_p]

        # # Change parameters
        self.camdll.set_parameter.restype = c_int
        self.camdll.set_parameter.argtypes = [c_char_p, c_float]

        # Load camera settings
        self.camdll.LoadCameraSettings.restype = c_int
        self.camdll.LoadCameraSettings.argtypes = [c_char_p]

        # Initialize and set dll ref
        self.camdll.Initialize()
        self.is_init = True

        # Load camera settings if path is specified
        if settings_path:
            self.camdll.LoadCameraSettings(settings_path.encode())

        # get camera info
        width = c_int(0)
        height = c_int(0)
        fps = c_float(0.)
        readout_time = c_float(0.)
        ecode = self.camdll.GetCameraInfo(byref(width), byref(height), byref(fps), byref(readout_time))
        self.check_for_error(ecode, '__init__')

        # init members
        self.width = width.value
        self.height = height.value
        self.fps = fps.value
        self.readout_time = readout_time.value * 1E-6

    def __del__(self):
        # Terminate
        if self.is_init:
            self.camdll.Terminate()

    def check_for_error(self, ecode, method):
        if ecode is not 0:
            raise Exception(f"PylonCamera: {method} failed, ecode: {ecode}")

    def set_cam_parameter(self, name, val):
        ecode = self.camdll.set_parameter(c_char_p(name.encode()), val)
        self.check_for_error(ecode, "Setting exposure time")

    def grab_image(self):
        buffer = (c_ubyte * (self.width * self.height))()
        ecode = self.camdll.GetImage(buffer, self.width, self.height)
        self.check_for_error(ecode, 'grab_image')

        img = np.reshape(buffer, (self.height, self.width))

        return img

    def grab_images_sequence(self, num_of_images):
        # create some arrays
        buffers = [np.zeros(self.width * self.height, dtype="ubyte") for ii in range(num_of_images)]

        # get ctypes handles
        ctypes_arrays = [np.ctypeslib.as_ctypes(array) for array in buffers]

        # Pack into pointer array
        pointer_ar = (POINTER(c_ubyte) * num_of_images)(*ctypes_arrays)

        ecode = self.camdll.GrabImagesSequence(num_of_images, pointer_ar)
        self.check_for_error(ecode, 'grab_images_sequence')

        for ii in range(num_of_images):
            buffers[ii] = np.reshape(buffers[ii], (self.height, self.width))

        return buffers

    def create_video(self, video_path, duration, fps):
        img_skipped = c_int(-1)
        ecode = self.camdll.CreateVideo(c_char_p(video_path.encode()), duration, fps, byref(img_skipped))
        self.check_for_error(ecode, 'create_video')

        print("%s create, %d img skipped" % (video_path, img_skipped.value))

    def save_camera_settings(self, filepath):
        ecode = self.camdll.SaveCameraSettings(filepath.encode())
        self.check_for_error(ecode, 'save_camera_settings')

    def load_camera_settings(self, filepath):
        ecode = self.camdll.LoadCameraSettings(filepath.encode())
        self.check_for_error(ecode, 'load_camera_settings')

    @CFUNCTYPE(None, c_int)
    def foo(self, arg):
        '''

        :return:
        '''
        print('foo Called with', arg)

    def find_centers_in_images_sequence(self, num_of_images):
        # create some arrays
        x = np.zeros(num_of_images, dtype="double")
        y = np.zeros(num_of_images, dtype="double")
        timestamp = np.zeros(num_of_images, dtype="long")

        # get ctypes handles
        x_ctypes = np.ctypeslib.as_ctypes(x)
        y_ctypes = np.ctypeslib.as_ctypes(y)
        timestamp_ctypes = np.ctypeslib.as_ctypes(timestamp)

        ecode = self.camdll.find_centers_in_images_sequence(int(self.width), int(self.height),
                                                            byref(timestamp_ctypes), byref(x_ctypes), byref(y_ctypes),
                                                            num_of_images, find_center)
        self.check_for_error(ecode, 'FindCentersInImagesSequence')

        return x, y, timestamp

    def grab_images_using_thread(self, num_of_images):
        # create some arrays
        x = np.zeros(num_of_images, dtype="double")
        y = np.zeros(num_of_images, dtype="double")
        timestamp = np.zeros(num_of_images, dtype="uint64")

        # get ctypes handles
        x_ctypes = np.ctypeslib.as_ctypes(x)
        y_ctypes = np.ctypeslib.as_ctypes(y)
        timestamp_ctypes = np.ctypeslib.as_ctypes(timestamp)

        # Blocking function to grab all centers
        ecode = self.camdll.grab_images_using_thread(int(self.width), int(self.height),
                                                        byref(timestamp_ctypes), byref(x_ctypes), byref(y_ctypes),
                                                     num_of_images, find_center)
        self.check_for_error(ecode, 'grab_images_using_thread')

        return x, y, timestamp



def set_obj_attr(obj, meta):
    for key in meta.keys():
        if type(meta[key]) is dict:
            obj.__setattr__(key, JsJson(meta[key]))
        elif type(meta[key]) is list:
            obj.__setattr__(key, [JsJson(item) for item in meta[key]])
        else:
            obj.__setattr__(key, meta[key])


class JsJson:
    def __init__(self, obj):
        set_obj_attr(self, obj)

    def __new__(cls, obj):
        if type(obj) is dict:
            ret = super(JsJson, cls).__new__(cls)
        elif type(obj) is list:
            ret = [JsJson(item) for item in obj]
        else:
            ret = obj

        return ret

    def as_dict_copy(self):
        out = self.__dict__.copy()
        for key in out.keys():
            if type(out[key]) is JsJson:
                out[key] = out[key].as_dict_copy()
            elif type(out[key]) is list:
                l = []
                for item in out[key]:
                    if type(item) is JsJson:
                        l.append(item.as_dict_copy())
                    else:
                        l.append(item)
                out[key] = l

        return out

    def __str__(self):
        return str(self.as_dict_copy())



import cv2
import os

from innolab_utils.pylon_camera_dll.PylonCamera import PylonCamera


def grab_image(pcam):
    img = pcam.grab_image()

    cv2.namedWindow('img', cv2.WINDOW_GUI_EXPANDED | cv2.WINDOW_FREERATIO)
    cv2.imshow('img', img)
    cv2.waitKey(0)
    cv2.destroyAllWindows()


def create_video(pcam):
    video_path = "dump/test_video.avi"
    pcam.create_video(video_path, 3, 25)

    vidcap = cv2.VideoCapture(video_path)
    framerate = vidcap.get(5)
    framecount = vidcap.get(7)

    print("framerate: %d, framecount: %d" % (framerate, framecount))


def grab_images_sequence(pcam):
    images = pcam.grab_images_sequence(130)

    cv2.namedWindow('img', cv2.WINDOW_GUI_EXPANDED | cv2.WINDOW_FREERATIO)

    for img in images:
        cv2.imshow('img', img)
        cv2.waitKey(3)

    cv2.destroyAllWindows()

    # dump images to video
    img = images[0]
    width = img.shape[1]
    height = img.shape[0]
    writer = cv2.VideoWriter("dump/grab_images_sequence.avi", cv2.VideoWriter_fourcc(*"XVID"), int(pcam.fps), (width, height), False)
    for img in images:
        writer.write(img)
    writer.release()


def save_and_load_camera_settings(pcam):
    filepath = './dump/save_and_load_camera_settings.pfs'
    if os.path.exists(filepath):
        os.remove(filepath)

    pcam.save_camera_settings(filepath)
    if not os.path.exists(filepath):
        raise("failed saving settings to %s" % filepath)

    pcam.load_camera_settings(filepath)


if __name__ == "__main__":
    path = r"C:\Development\feedback_calibration\src\bgVGA.pfs"
    pcam = PylonCamera(path)

    grab_image(pcam)

    # create_video(pcam)
    # grab_images_sequence(pcam)
    # save_and_load_camera_settings(pcam)



import logging

import colorlog


def init_logger(name, level=logging.INFO, log_path='report.log', console_enable=True):
    '''
    base on https://docs.python.org/3/howto/logging-cookbook.html
    :param name: logger name
    :param level: logging level
    :param log_path: log file handler default path
    :return:
    '''
    # create logger with 'spam_application'
    logger = logging.getLogger(name)
    logger.setLevel(level)

    # create file handler which logs even debug messages
    fh = logging.FileHandler(log_path)
    fh.setLevel(level)

    # set file formatter
    formatter = logging.Formatter('[%(asctime)s] [%(name)s] [%(levelname)s] %(message)s')
    fh.setFormatter(formatter)
    logger.addHandler(fh)

    if console_enable:
        # create console handler with a higher log level
        ch = colorlog.StreamHandler()
        ch.setLevel(level)

        # set console formatter
        formatter = colorlog.ColoredFormatter('%(log_color)s[%(asctime)s] [%(name)s] [%(levelname)s] %(message)s')
        ch.setFormatter(formatter)
        logger.addHandler(ch)

    return logger


import asyncio
import numpy as np
import math
import time
import matplotlib.pyplot as plt

from innopy.api import DeviceInterface, DeviceType, logger_init
from innolab_utils.run_board_api.TapHandler import TapHandler

from logger_config import logger


class RunBoardApi:
    __MAX_NUM_OF_MIRRORS__ = 4

    def __init__(self, config, num_of_mirrors=1, outpath=None):
        '''
        Connect to board and sets board parameters.
        Sens tap is enabled by default
        :param config: contains all the configurations of the device
        :param num_of_mirrors: number of mirror activated
        '''

        self.tap_handler = None
        self.di = None
        self.outpath = outpath
        self.num_of_mirrors = num_of_mirrors
        self.ip = config.ip
        # set api logger
        date = time.strftime('%Y-%m-%d', time.localtime(time.time()))
        log_filename = f"api_logs/{self.ip}_{date}.txt"
        logger_init(log_filename, 0)

        # Configure the test of the run board and transfer it to the Tap Handler
        self.test = config.main_config.test

        if num_of_mirrors > self.__MAX_NUM_OF_MIRRORS__:
            raise Exception(f"num_of_mirrors '{num_of_mirrors}' exceed max value '{self.__MAX_NUM_OF_MIRRORS__}'")

        # BSample parameters defaults
        if config.Model.Type == "B":
            self.b2 = config.Model.BSample.b
            self.a2 = config.Model.BSample.a
            self.a_led = config.Model.BSample.a_led
            self.b_led = config.Model.BSample.b_led
        # CSample parameters defaults
        else:
            self.b2 = config.Model.CSample.b
            self.a2 = config.Model.CSample.a
            self.a_led = config.Model.CSample.a_led
            self.b_led = config.Model.CSample.b_led

        # board parameters initialization sequence
        self.di = DeviceInterface(ip=self.ip, password="200andbeyond", login_level=2)

        self.di.set_dp("DriveSize", config.MEMsRunBoard.table_size)
        self.di.set_dp("BuffersAfterStop", config.MEMsRunBoard.BuffersAfterStop)
        self.di.set_dp("SamplingRate", float(config.MEMsRunBoard.rate))
        self.di.set_dp("ZeroCode", int(config.MEMsRunBoard.zero_daq))
        self.di.set_dp("MaxPulse", float(config.MEMsRunBoard.trigger_pulse_width) * np.ones(4, np.float32))

        if config.Model.Type == "B":
            self.di.set_dp("LEDDrive", float(config.MEMsRunBoard.led_drive))
        else:
            led_dac = np.uint16(
                (1000 * float(config.MEMsRunBoard.led_drive) * self.a_led + self.b_led) * (2 ** 16 - 1) / 2.5)
            self.di.set_dp("LEDDriveDAC", led_dac)

        self.stop()
        self.di.set_dp("SSTime", float(config.MEMsRunBoard.raster_SSTime))
        self.di.set_dp("IntPower", int(config.MEMsRunBoard.raster_int_power))

        self.drive_size = config.MEMsRunBoard.table_size
        self.sense_cycle_size = 0

    @property
    def cycle_size(self):
        '''

        :return:
        '''
        return self.tap_handler.cycle_size

    @property
    def tap_count(self):
        return self.tap_handler.count

    async def module_temp(self):
        '''
        Getting the temperature from the Module_sample
        :return:
        '''
        self.di.set_dp("ReadTempInterval", 0.5)  # turn On the temperature sampling
        await asyncio.sleep(2)
        temp = self.di.get_dp("Temperature")
        self.di.set_dp("ReadTempInterval", -1.0)  # turn Off the temperature sampling
        return temp

    @property
    def spaceship_temp(self):
        '''
        Getting the temperature from the spaceship
        :return:
        '''
        return self.di.get_dp("spaceship_temp")

    @property
    def eeprom(self):
        '''
        Returns all data in Eporm as object (parsed)
        :return:
        '''
        return self.di.get_dp("eeprom")

    @property
    def spaceship_humidity(self):
        '''

        :return:
        '''
        return self.di.get_dp("spaceship_humidity")

    @property
    def tap_size(self):
        return self.tap_handler.sense_buff_size

    def start(self):
        '''
        starts drive and sampling ( for 1 mirror or hole spaceship ? )
        :return:
        '''
        self.di.set_dp("SampleActive", int(True))

    # async def stop(self):
    def stop(self):
        '''
        stops drive and sampling ( for 1 mirror or hole spaceship ? )
        :return:
        '''
        self.di.set_dp("SampleActive", int(False))
        # todo temp in comment, need to update embedded version
        # Letting the device stop properly
        # while self.di.get_dp("active") != 0:
        #     await asyncio.sleep(0.01)

    def init_tap_handler(self):
        self.tap_handler = TapHandler(self.di, self.test, self.num_of_mirrors, self.ip)
        # reset counter of the tap handler
        self.tap_handler.reset_counter()

    def set_tap_size(self, size):
        self.tap_handler.set_cycle_size(size)

    async def del_tap_handler(self):
        if self.tap_handler:
            logger.info("del_tap_handler : Start")  # todo temp
            self.tap_handler.unregister()
            # self.tap_handler.reset_counter(reset_tap_index=True)
            await asyncio.sleep(1) # wait for all callbacks
            self.tap_handler.del_di()
            self.tap_handler = None
            logger.info("del_tap_handler : Done")

    # def continue_tap_handler(self):
    #     # self.ignore_first_taps()
    #     self.tap_handler.reset_counter(reset_tap_index=True)
    #     self.tap_handler.register()
    #
    # def stop_tap_handler(self):
    #     self.tap_handler.unregister()
    #     self.tap_handler.reset_counter(reset_tap_index=True)
    #     # self.clear_tap_buffer()

    def set_drive_volt(self, volts):
        '''
        Setting the drive. drive size
        :param volts: volts is a numpy array in size NxM, where N(default=4) is the number of dac per mirror
        :return:
        '''
        dword = self.di.get_empty_dp("DriveTable")

        # converting the Volts to DACWords
        for i, vals in enumerate(volts):
            for j, v in enumerate(vals):
                dword[i][j] = math.floor((v - self.b2) * (2 ** 16 - 1) / (self.a2 * 2.5))
        self.set_drive_dword(dword)

    def set_drive_dword(self, dword):
        '''
        Setting the drive. drive size
        :param dword: dword is a numpy array in size NxM, where N(default=4) is the number of dac per mirror
        :return:
        '''
        dp = self.di.get_empty_dp("DriveTable")
        if dp.shape == dword.shape:
            self.di.set_dp("DriveTable", np.array(dword, np.uint16))
        else:
            raise Exception(f'The Drive table is not in the correct shape, {dp.shape()} != {dword.shape()}')

    def check_error(self):
        '''
        @todo close interface with System. pulling? event(tap)?
        Checking if one of the mirrors has an error
        :return:
        '''
        raise NotImplemented()

    def check_warning(self):
        '''
        @todo close interface with System. pulling? event(tap)?
        Checking if one of the mirrors has an error
        :return:
        '''
        raise NotImplemented()

    def tap_start(self, name):
        self.di.set_tap_activation_state(name, True)

    def tap_stop(self, name):
        self.di.set_tap_activation_state(name, False)

    def set_ss_time(self, val):
        self.di.set_dp("SSTime", val)

    def set_int_power(self, val):
        self.di.set_dp("IntPower", val)

    def set_dp(self, name, val, set_to_flash=False):
        self.di.set_dp(name, val, set_param=set_to_flash)

    def get_dp(self, name):
        return self.di.get_dp(name)

    def get_empty_dp(self, name):
        return self.di.get_empty_dp(name)

    def get_aux_cycle(self):
        return self.tap_handler.get_aux_buffer()

    def get_cycle(self):
        # waiting until all the buffer is full
        while not self.tap_handler.check_buffer:
            time.sleep(0.05)

        return self.tap_handler.get_buffer()

    def ignore_first_taps(self):
        self.tap_handler.set_ignore_taps()

    def clear_tap_buffer(self):
        self.tap_handler.clear_buffer_full()

    async def get_async_cycle(self):
        # waiting until all the buffer is full
        valid, data = await self.tap_handler.get_buffer()
        while not valid:
            await asyncio.sleep(0.05)
            valid, data = await self.tap_handler.get_buffer()
        return data

    async def reset(self, wait_after_reset=True):
        try:
            self.di.set_dp("reset", 1)
            if wait_after_reset:
                await asyncio.sleep(10)
        except Exception as e:
            logger.error(f"[{self.ip}][RunBoardApi: reset] error during reset {e}")

    def __del__(self):
        logger.info("RunboardApi destructor Start")
        # if self.tap_handler: todo needed?
        #     self.del_tap_handler()
        self.di = None
        logger.info("RunboardApi destructor Done")
    # #     self.tap_handler.unregister()




import numpy as np
import time
import traceback
from logger_config import logger

class TapHandler:
    __TAP_TO_IGNORE__ = 2
    __MAX_PREMA_SENSORS__ = 4
    counter = 0

    def __init__(self, di, test, num_of_mirrors=1, ip='10.1.1.112'):
        self.is_done = False
        self.num_of_mirrors = num_of_mirrors
        self.di = di
        self.int_power = int(di.get_dp("IntPower"))
        self.test = test
        if self.test == "CALIBRATION":
            self.buf_size = 1
        else:
            self.buf_size = 2

        # for debug prints
        self.last_callback = time.time()
        self.current_callback = time.time()
        self.timeout = 3  # 3 seconds
        self.ip = ip

        self._count = np.zeros([self.buf_size], dtype=np.int32)
        self.__NUM_OF_BUFFERS__ = self.buf_size

        # Sense size should contain 1 cycle
        self._sense_cycle_size = int(di.get_dp('DriveSize') * (2 ** self.int_power))

        # To concatenate the data
        self._sense_buff = np.zeros([num_of_mirrors, self.__MAX_PREMA_SENSORS__, self._sense_cycle_size,
                                     self.__NUM_OF_BUFFERS__], np.uint16)
        # TO accumulate the current taps
        self.tap_buffer = np.zeros([num_of_mirrors, (di.get_empty_dp('sense0')).shape[0],
                                    (di.get_empty_dp('sense0')).shape[1]], np.uint16)

        self._buffer_index = 0  # starting to fill the buffer from this index
        self.is_buf_locked = np.zeros(self.buf_size, dtype=bool)
        self.is_buffer_full = np.zeros(self.buf_size, dtype=bool)

        self.frame_index = 0
        self.tap_index = 0
        self.wrong_frame = 0  # ignore first tap

        self._sense_default_len = len(di.get_empty_dp('sense0')[0])
        self.num_of_subframes = int(np.ceil(self._sense_cycle_size / self._sense_default_len))
        self.last_subframe_len = self._sense_cycle_size % self._sense_default_len
        logger.info(f"table_size:{self._sense_cycle_size}, sense_len:{self._sense_default_len}, num_of_subframes {self.num_of_subframes}")

        self.is_stop_request = False

        # Register event
        self.is_registered = False
        self.register()

    def reset_counter(self, index=-1, reset_tap_index =False):
        '''
        Resetting one of the indexes or Both of the double buffer
        :param index:
        :return:
        '''

        if index == 0 or index == 1:
            self._count[index] = 0
        else:
            # reset all of the counters
            self._count[:] = 0

        if reset_tap_index:
            self.frame_index = 0
            self.tap_index = 0

    def set_cycle_size(self, int_power):
        self._sense_cycle_size = int(self.di.get_dp('DriveSize') * (2 ** int_power))

    @property
    def cycle_size(self):
        return self._sense_cycle_size

    @property
    def check_buffer(self):
        logger.info(f"buffer in index {self._buffer_index} is {self.is_buffer_full[self._buffer_index]}")
        return self.is_buffer_full[self._buffer_index]

    async def get_buffer(self):
        '''
        If the buffer index is 1 meaning buffer is full at o index and the opposite
        :return: sense_buff
        '''
        i = self._buffer_index 
        # i = (self._buffer_index + 1) % self.buf_size
        # Waiting for this buffer to be full ( if full the count is set to 0 )
        # while self._count[i] != 0:
        #     time.sleep(0.05)
        #     # TODO: ask Maxim
        if not self.is_buffer_full[i]:
            return False, -1

        # Locking the buffer not to be inserted in the middle
        self.is_buf_locked[i] = True

        # Extracting the buffer
        ret = self._sense_buff[:, :, :, i]
        self.is_buffer_full[i] = False

        # Releasing the buffer
        self.is_buf_locked[i] = False

        logger.debug(f'get_buffer[{i}]:: returning the buffer')
        return True, ret
        # return 1, ret

    def set_ignore_taps(self):
        self.__TAP_TO_IGNORE__ = 2

    # def set_stop_request(self, val):
    #     self.is_stop_request = val
    #     self.frame_index = 0
    #     self.tap_index = 0

    def clear_buffer_full(self):
        self.is_buffer_full[self._buffer_index] = False

    def set_wrong_frame(self, msg):
        logger.warning(msg)
        self.wrong_frame = self.frame_index

    def callback(self, handler, count):
        '''
        Filling the double buffer in turns
        After one buffer is full we set the boolean array buffer_full to True and start filling the other one
        :param handler:
        :param count:
        :return:
        '''

        logger.debug(f"{[self.ip]} Tap callback Start")

        # if self.test == "ENDURANCE" and self.__TAP_TO_IGNORE__ > 0:
        #     logger.debug(f"self.__TAP_TO_IGNORE__: {self.__TAP_TO_IGNORE__}")
        #     self.__TAP_TO_IGNORE__ -= 1
        #     return

        # if self.is_stop_request:
        #     logger.info(f"{[self.ip]}[Tap callback] during stop request, ignore these taps")
        #     return

        if self.test == "CALIBRATION" and self.is_done:
            return

        try:
            # in case the interrupt accrues while we grab the buffer
            while self.is_buf_locked[self._buffer_index]:
                logger.warning(f"Buffer is blocked waiting 3 ms")
                time.sleep(3 / 1000)  # sleep of 3 ms

            # debug for the "CALIBRATION" test
            if self.test == "CALIBRATION" and self._count[0] % 250 == 0:
                logger.debug(f'taps collected so far: {self._count[0]} out of {self._sense_cycle_size}')

            # Get tap event and fill buffer with sense data
            names = self.di.get_taps_callback_dps(handler, count)

            last_frame_index = self.frame_index
            last_tap_index = self.tap_index

            if "frameCnt" in names:
                self.frame_index = self.di.get_dp("frameCnt")
                logger.debug(f"{[self.ip]}[callback] frame_index {self.frame_index}")
            else:
                self.set_wrong_frame("frameCnt tap is missing")

            if "tapCnt" in names:
                self.tap_index = self.di.get_dp("tapCnt")
                logger.debug(f"{[self.ip]}[callback] tapCnt {self.tap_index}")
            else:
                self.set_wrong_frame("tapCnt tap is missing")

            for i in range(self.num_of_mirrors):
                if f"sense{i}" in names:
                    sense = self.di.get_dp(f"sense{i}")
                    self.tap_buffer[i, :, :] = sense
                else:
                    self.set_wrong_frame(f"sense{i} is missing")

            if self.frame_index < last_frame_index:
                # raise Exception(f"{[self.ip]}[callback] bad frame index: {last_frame_index}->{self.frame_index}")
                self.set_wrong_frame(f"{[self.ip]}[callback] bad frame index: {last_frame_index}->{self.frame_index}")


            if self.tap_index > self.num_of_subframes:
                self.set_wrong_frame(f"{[self.ip]}[callback] tapCnt > {self.num_of_subframes} ({self.tap_index}), at frame {self.frame_index}")

            if last_frame_index == self.frame_index:
                if self.tap_index != last_tap_index + 1:
                    self.set_wrong_frame(f"{[self.ip]}[callback] bad tap index: {last_tap_index}->{self.tap_index}"
                                         f" at frame {self.frame_index}")

            elif self.tap_index != 0 or last_tap_index != 3:
                self.set_wrong_frame(f"{[self.ip]}[callback] missed taps: {last_tap_index}->{self.tap_index}"
                                     f" from frame {last_frame_index}->{self.frame_index}")

            if self.frame_index == self.wrong_frame:
                logger.warning(f"skip frame {self.frame_index}")
                self.reset_counter(self._buffer_index)
                return

            # If cycle size is smaller than len, use cycle size
            size = min(self._sense_cycle_size, self._sense_default_len)

            # Blocking this buffer index until we finish to fill it
            self.is_buf_locked[self._buffer_index] = True

            cur_count = self._count[self._buffer_index]

            # Last accumulation
            if cur_count + (self.tap_buffer[0]).shape[1] > self._sense_cycle_size:
                last_size = self._sense_cycle_size - cur_count

                if last_size != self.last_subframe_len:
                    logger.error(f'wrong accumulation {last_size} != {self.last_subframe_len}')

                self._sense_buff[:, :, cur_count:, self._buffer_index] \
                    = self.tap_buffer[:, :, :last_size]

                # Setting the buffer flag to true
                self.is_buffer_full[self._buffer_index] = True

                logger.debug(f'taps collected so far[{self._buffer_index}]: {self._count[self._buffer_index]+last_size}'
                             f' out of {self._sense_cycle_size}')

                # Resetting the buffer counter of the full one
                self.reset_counter(self._buffer_index)

                # Switching the buffer index to fill the next one
                # to get 0/1 only or just 1 in Calibration
                self._buffer_index = (self._buffer_index + 1) % self.buf_size

                # for the CALIBRATION test
                self.is_done = True

            # First accumulation 0:1023
            else:
                if cur_count != (size * self.tap_index):
                    logger.error(f"Wrong accumulation {cur_count} != {size * self.tap_index}")

                cur_size = cur_count+size
                self._sense_buff[:, :, cur_count:cur_size, self._buffer_index] = \
                    self.tap_buffer[:, :, :size]

                self._count[self._buffer_index] += size
                logger.debug(f'taps collected so far[{self._buffer_index}]: {self._count[self._buffer_index]} '
                             f'out of {self._sense_cycle_size}')

            # Release the buffer
            self.is_buf_locked[self._buffer_index] = False

            logger.debug(f"{[self.ip]} Tap callback Done")

        except Exception as e:
            self.reset_counter(self._buffer_index) # todo is it ok for all cases
            self.wrong_frame = self.frame_index
            logger.info(f'{[self.ip]} TapHandler: failed executing callback. reason: {e}')
            logger.info(f' names: {names}')
            logger.info(f'traceback: {traceback._cause_message}')
            traceback.print_exc()
            # raise Exception(f'TapHandler: failed executing callback. reason: {e}')

    def register(self):
        if not self.is_registered:
            self.di.register_taps_callback(self.callback)
            logger.info(f"{[self.ip]} Tap callback register")
            self.is_registered = True
        else:
            logger.info(f"{[self.ip]} Tap callback is already registered")

    def unregister(self):
        if self.is_registered:
            self.di.unregister_taps_callback()
            logger.info(f"{[self.ip]} Tap callback unregister")
            self.is_registered = False
        else:
            logger.info(f"{[self.ip]} Tap callback is already unregistered")

    def del_di(self):
        self.di = None

    # def __del__(self):
    #     logger.info("## TapHandler destructor ##")
    #     self.unregister()






import json

from innoviz_lib.jsjson import JsJson

example_path = './example.json'

with open(example_path, 'r') as f:
    meta = json.load(f)

out = JsJson(meta)

print(out)
print("out.a : %d" % out.a)
print("out.d.b : %.2f" % out.d.b)




from innopy.api import *
from src.json_config import get_config


def main():
    ip = "10.1.1.101"
    di = DeviceInterface()

    di.set_dp("networkIPAddr", ip, True)


if __name__ == "__main__":
    main()



import numpy as np

from innolab_utils.pylon_camera_dll.Test.traces_gui import IndexGUI


file_path = "C:\\Users\\maxim.kuzminsky\\Desktop\\spot_test.csv"
app = IndexGUI(file_path)
app.root.mainloop()

